/* directionTable.h
 */
#ifndef BOARD_TABLE_H
#define BOARD_TABLE_H

#include "osl/direction.h"
#include "osl/misc/carray.h"
#include "osl/offset32.h"

namespace osl
{
  class BoardTable
  {
    CArray<Direction,Offset32::SIZE> directions;
    CArray<Offset, Offset32::SIZE> short_offsets;
    CArray<Offset, Offset32::SIZE> short_offsets_not_knight;
#ifndef MINIMAL
    CArray<int,Offset32Wide::SIZE> space_counts;
#endif
  public:
    static const CArray<Offset, DIRECTION_SIZE> offsets;
    static const CArray<int, DIRECTION_SIZE> dxs;
    static const CArray<int, DIRECTION_SIZE> dys;
  private:
    template<Direction Dir>
    void setDirections();
    template<Direction Dir>
    void setKnightDirections();
    void init();
  public:
    /**
     * 黒にとってのoffsetを返す
     */
    const Offset getOffsetForBlack(Direction dir) const{
      return offsets[static_cast<int>(dir)];
    }
    int getDxForBlack(Direction dir) const{
      return dxs[static_cast<int>(dir)];
    }
    int getDyForBlack(Direction dir) const{
      return dys[static_cast<int>(dir)];
    }
    template<Player P>
    const Offset getOffset(Direction dir) const{
      return getOffsetForBlack(dir)*playerToMul(P);
    }
    const Offset getOffset(Player pl,Direction dir) const{
      if (pl==BLACK)
	return getOffset<BLACK>(dir);
      else
	return getOffset<WHITE>(dir);
    }

    /**
     * next position from pos for player P.
     * 答えが isOnBoard とは限らない
     */
    const Position nextPosition(Player P, Position pos, Direction dr) const
    {
      assert(pos.isOnBoard());
      const Offset offset = getOffset(P, dr);
      return pos + offset;
    }

    BoardTable();
    /** @param P どちらのPlayerにとっての方向かを指定 */
    template <Player P>
    Direction getLongDirection(Offset32 offset32) const
    {
      assert(offset32.isValid());
      const Offset32 blackOffset32 = offset32.blackOffset32<P>();
      Direction ret=directions[blackOffset32.index()];
      assert(isValid(ret));
      return ret;
    }
    Direction getLongDirection(Player P, Offset32 offset32) const
    {
      if (P == BLACK) 
	return getLongDirection<BLACK>(offset32);
      else
	return getLongDirection<WHITE>(offset32);
    }
    /** @param P どちらのPlayerにとっての方向かを指定 */
    template <Player P>
    Direction getLongDirection(Position from, Position to) const
    {
      return getLongDirection<P>(Offset32(to,from));
    }
#ifndef MINIMAL
    /**
     * fromとtoが長い利きを持つ位置にある時，間のマスの数を求める
     * 一致している時は0
     * となりも0
     * 関係ない時は-1
     */
    int spaceCounts(Position from,Position to) const
    {
      Offset32Wide offset32(from,to);
      return space_counts[offset32.index()];
    }
#endif
    /**
     * Longの利きの可能性のあるoffsetの場合は, 反復に使う offsetを
     * Shortの利きのoffsetの場合はそれ自身を返す.
     * 利きの可能性のない場合は0を返す
     */
    const Offset getShortOffset(Offset32 offset32) const{
      assert(offset32.isValid());
      return short_offsets[offset32.index()];
    }
    /**
     * Longの利きの可能性のあるoffsetの場合は, 反復に使う offsetを
     * Knight以外のShortの利きのoffsetの場合はそれ自身を返す.
     * Knightの利き, 利きの可能性のない場合は0を返す
     */
    const Offset getShortOffsetNotKnight(Offset32 offset32) const{
      assert(offset32.isValid());
      return short_offsets_not_knight[offset32.index()];
    }
    /**
     * p0, p1の間にtがあるかどうか.
     * (t,p0), (t,p1)のどちらかが8方向である時にのみ呼び出すこと
     */
    bool isBetween(Position t,Position p0,Position p1) const
    {
      Offset offset0=getShortOffsetNotKnight(Offset32(p0,t));
      Offset offset1=getShortOffsetNotKnight(Offset32(p1,t));
      assert(!offset0.zero() || !offset1.zero());
      return (-offset0)==offset1;
    }
  };

  extern const BoardTable Board_Table;
} // namespace osl


#endif /* BOARD_TABLE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
